from .gradient_image_or_color import Color


def consider_points_equals(point1: tuple[float, float], point2: tuple[float, float]) -> bool:
    return point1 * point1 + point2 * point2 < 0.001


def hex_to_color(hex_code: str, opacity: float = 1.0) -> Color:
    hex_code = hex_code.lstrip('#')
    return Color(
        int(hex_code[0:2], 16) / 255,
        int(hex_code[2:4], 16) / 255,
        int(hex_code[4:6], 16) / 255,
        opacity,
    )



from copy import deepcopy
import json
from typing import Callable

from .gradient_image_or_color import (
    GradientImageOrColor,
    Color,
    GradientStop,
    LinearGradient,
    RadialGradient,
    Image,
)
from .utils import consider_points_equals


class VectorObject:
    def __init__(
        self,
        scene,
        points: list[tuple[float, float]] = None,
        fill: GradientImageOrColor | None = None,
        fill_rule: str = 'nonzero',
        stroke: GradientImageOrColor | None = None,
        stroke_width: float = 0.0,
        line_cap: str = 'butt',
        line_join: str = 'miter',
        subobjects: list['VectorObject'] = None,
        index: int = 0,
    ) -> None:
        self.scene = scene
        self.points = points or []
        self.fill = fill or GradientImageOrColor.default()
        self.fill_rule = fill_rule
        self.stroke = stroke or GradientImageOrColor.default()
        self.stroke_width = stroke_width
        self.line_cap = line_cap
        self.line_join = line_join
        self.subobjects = subobjects or []
        self.index = index
    
    async def create(self, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'createVectorObject', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def draw_stroke_then_fill(self, t: float, default_stroke_width: float = 2.0) -> "VectorObject":
        data = json.dumps([self.to_dict(), t, default_stroke_width])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'drawStrokeThenFill', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def fade_in(self, scale_factor: float, shift: tuple[float, float], t: float) -> "VectorObject":
        shift = [shift[0], shift[1]]
        data = json.dumps([self.to_dict(), scale_factor, shift, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'fadeIn', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def fade_out(self, scale_factor: float, shift: tuple[float, float], t: float) -> "VectorObject":
        shift = [shift[0], shift[1]]
        data = json.dumps([self.to_dict(), scale_factor, shift, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'fadeOut', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def grow_arrow_with_final_tip(self, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'growArrowWithFinalTip', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def grow_arrow_with_initial_tip(self, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'growArrowWithInitialTip', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def grow_arrow_with_tips_at_both_ends(self, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'growArrowWithTipsAtBothEnds', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def grow_from_center(self, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'growFromCenter', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def morph_shape(self, target: "VectorObject", t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), target.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'morphShape', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def rotate_animation(self, angle: float, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), angle, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotateAnimation', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def scale_in_place(self, scale_factor: float, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), scale_factor, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'scaleInPlace', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def set_fill_animation(self, color: Color, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), color.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'setFillAnimation', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def set_stroke_animation(self, color: Color, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), color.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'setStrokeAnimation', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def shift_animation(self, shift: tuple[float, float], t: float) -> "VectorObject":
        shift = [shift[0], shift[1]]
        data = json.dumps([self.to_dict(), shift, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'shiftAnimation', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def show_temporarily(self, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'showTemporarily', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def spinning_grow(self, angle: float, t: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), angle, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'spinningGrow', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    def get_index(self) -> int:
        return self.index
    
    def set_index(self, index: int) -> "VectorObject":
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=self.subobjects,
            index=index,
        )
    
    async def increment_index(self, increment: int, recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), increment, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'incrementIndex', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    def get_points(self) -> list[tuple[float, float]]:
        return self.points
    
    def set_points(self, points: list[tuple[float, float]]) -> "VectorObject":
        return VectorObject(
            self.scene,
            points=points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=self.subobjects,
            index=self.index,
        )
    
    def set_subobjects(self, subobjects: list['VectorObject']) -> "VectorObject":
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def get_subobjects(self) -> list['VectorObject']:
        return self.subobjects
    
    async def get_subobjects_recursively(self, with_points: bool) -> list['VectorObject']:
        data = json.dumps([self.to_dict(), with_points])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getSubobjectsRecursively', {data})")
        return [VectorObject.from_dict(self.scene, subobject) for subobject in result]
    
    def add(self, subobject: "VectorObject") -> "VectorObject":
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=self.subobjects + [subobject],
            index=self.index,
        )
    
    def remove(self, i: int) -> "VectorObject":
        subobjects = self.subobjects.copy()
        subobjects.pop(i)
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def get_subobject(self, i: int) -> "VectorObject":
        return self.subobjects[i]
    
    def set_subobject(self, i: int, subobject: "VectorObject") -> "VectorObject":
        subobjects = self.subobjects.copy()
        subobjects[i] = subobject
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def slice_subobjects(self, start: int, end: int) -> list["VectorObject"]:
        return self.subobjects[start:end]
    
    def set_slice_subobjects(self, start: int, end: int, subobjects: list["VectorObject"]) -> "VectorObject":
        subobjects = self.subobjects[:start] + subobjects + self.subobjects[end:]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    async def get_subcurve(self, start: float, end: float) -> "VectorObject":
        data = json.dumps([self.to_dict(), start, end])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getSubcurve', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def get_partial_copy(self, start: float, end: float, recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), start, end, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getPartialCopy', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def get_anchors_and_handles(self) -> tuple[list[tuple[float, float]], list[tuple[float, float]], list[tuple[float, float]], list[tuple[float, float]]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getAnchorsAndHandles', {data})")
        return (
            [(point[0], point[1]) for point in result[0]],
            [(point[0], point[1]) for point in result[1]],
            [(point[0], point[1]) for point in result[2]],
            [(point[0], point[1]) for point in result[3]],
        )
    
    async def scale_handle_to_anchor_distances(self, scale: float, recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), scale, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'scaleHandleToAnchorDistances', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def set_anchors_and_handles(
        self,
        anchors_and_handles: tuple[list[tuple[float, float]], list[tuple[float, float]], list[tuple[float, float]], list[tuple[float, float]]]
    ) -> "VectorObject":
        anchors_and_handles = [
            [[point[0], point[1]] for point in anchors_and_handles[0]],
            [[point[0], point[1]] for point in anchors_and_handles[1]],
            [[point[0], point[1]] for point in anchors_and_handles[2]],
            [[point[0], point[1]] for point in anchors_and_handles[3]],
        ]
        data = json.dumps([self.to_dict(), anchors_and_handles])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'setAnchorsAndHandles', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def get_nth_curve_points(self, n: int) -> list[tuple[float, float]]:
        data = json.dumps([self.to_dict(), n])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getNthCurvePoints', {data})")
        return [(point[0], point[1]) for point in result]
    
    async def get_nth_curve_length_pieces(self, n: int, sample_points: int = 10) -> list[float]:
        data = json.dumps([self.to_dict(), n, sample_points])
        return await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getNthCurveLengthPieces', {data})")
    
    def get_num_curves(self) -> int:
        return len(self.subobjects) // 4
    
    def is_closed(self) -> bool:
        return consider_points_equals(self.points[0], self.points[-1])
    
    async def get_subpaths(self) -> list[list[tuple[float, float]]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getSubpaths', {data})")
        return [[(point[0], point[1]) for point in subpath] for subpath in result]
    
    async def apply_function(
        self,
        function: Callable[[float, float], tuple[float, float]],
        recursive: bool,
        about_point: tuple[float, float] | None = None,
        about_edge: tuple[float, float] | None = None,
    ) -> "VectorObject":
        id_ = self.scene.register_callback(function)
        data = json.dumps([self.to_dict(), id_, recursive, about_point, about_edge])
        result = await self.scene.client.run_javascript(f"return await runMethod({self.scene.id}, 'applyFunction', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def get_pieces(self, n_pieces: int) -> "VectorObject":
        data = json.dumps([self.to_dict(), n_pieces])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getPieces', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def get_cubic_bezier_tuples(self) -> list[tuple[tuple[float, float], tuple[float, float], tuple[float, float], tuple[float, float]]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getCubicBezierTuples', {data})")
        return [
            (
                (point[0], point[1]),
                (point[2], point[3]),
                (point[4], point[5]),
                (point[6], point[7]),
            )
            for point in result
        ]
    
    async def scale(self, scale_factor: float, recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), scale_factor, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'scale', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def stretch(self, stretch_factor: tuple[float, float], recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), [stretch_factor[0], stretch_factor[1]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'stretch', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def shift(self, shift: tuple[float, float], recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), [shift[0], shift[1]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'shift', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def merged_points(self) -> list[tuple[float, float]]:
        data = json.dumps([self.to_dict()])
        return await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'mergedPoints', {data})")
    
    async def get_bounding_box(self) -> tuple[tuple[float, float], tuple[float, float]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getBoundingBox', {data})")
        return ((result[0][0], result[0][1]), (result[1][0], result[1][1]))
    
    async def get_center(self) -> tuple[float, float]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getCenter', {data})")
        return (result[0], result[1])
    
    async def get_center_of_mass(self) -> tuple[float, float]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getCenterOfMass', {data})")
        return (result[0], result[1])
    
    async def get_height(self) -> float:
        data = json.dumps([self.to_dict()])
        return await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getHeight', {data})")
    
    async def get_width(self) -> float:
        data = json.dumps([self.to_dict()])
        return await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getWidth', {data})")
    
    def get_fill_opacity(self) -> float:
        return (
            self.fill.gradient_image_or_color.a
            if isinstance(self.fill.gradient_image_or_color, Color)
            else self.fill.gradient_image_or_color.alpha
        )
    
    def set_fill_opacity(self, opacity: float, recursive: bool) -> "VectorObject":
        fill = self.fill.gradient_image_or_color
        if isinstance(fill, Color):
            new_fill = Color(fill.r, fill.g, fill.b, a=opacity)
        elif isinstance(fill, LinearGradient):
            new_stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in fill.stops
            ]
            new_fill = LinearGradient(fill.start, fill.end, new_stops, alpha=opacity)
        elif isinstance(fill, RadialGradient):
            new_stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in fill.stops
            ]
            new_fill = RadialGradient(fill.c, fill.r, fill.f, new_stops, alpha=opacity)
        else:
            new_fill = Image(fill.image_base64, fill.mime_type, fill.top_left, fill.bottom_right, alpha=opacity)
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_fill_opacity(opacity, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=GradientImageOrColor(new_fill),
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def set_stroke_opacity(self, opacity: float, recursive: bool) -> "VectorObject":
        stroke = self.stroke.gradient_image_or_color
        if isinstance(stroke, Color):
            new_stroke = Color(stroke.r, stroke.g, stroke.b, a=opacity)
        elif isinstance(stroke, LinearGradient):
            new_stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in stroke.stops
            ]
            new_stroke = LinearGradient(stroke.start, stroke.end, new_stops, alpha=opacity)
        elif isinstance(stroke, RadialGradient):
            new_stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in stroke.stops
            ]
            new_stroke = RadialGradient(stroke.c, stroke.r, stroke.f, new_stops, alpha=opacity)
        else:
            new_stroke = Image(stroke.image_base64, stroke.mime_type, stroke.top_left, stroke.bottom_right, alpha=opacity)
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_stroke_opacity(opacity, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=GradientImageOrColor(new_stroke),
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )

    async def move_to(self, point: tuple[float, float], recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), [point[0], point[1]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'moveTo', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def rotate(self, angle: float, recursive: bool) -> "VectorObject":
        data = json.dumps([self.to_dict(), angle, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotate', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def next_to_other(
        self,
        other: "VectorObject",
        direction: tuple[float, float],
        buff: float,
        aligned_edge: tuple[float, float],
        recursive: bool,
    ) -> "VectorObject":
        data = json.dumps([self.to_dict(), other.to_dict(), [direction[0], direction[1]], buff, [aligned_edge[0], aligned_edge[1]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'nextToOther', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def next_to_point(
        self,
        point: tuple[float, float],
        direction: tuple[float, float],
        buff: float,
        aligned_edge: tuple[float, float],
        recursive: bool,
    ) -> "VectorObject":
        data = json.dumps([self.to_dict(), [point[0], point[1]], [direction[0], direction[1]], buff, [aligned_edge[0], aligned_edge[1]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'nextToPoint', {data})")
        return VectorObject.from_dict(self.scene, result)
    
    async def arrange_subobjects(
        self,
        direction: tuple[float, float],
        buff: float,
        aligned_edge: tuple[float, float],
        recursive: bool
    ) -> "VectorObject":
        data = json.dumps([self.to_dict(), [direction[0], direction[1]], buff, [aligned_edge[0], aligned_edge[1]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'arrangeSubobjects', {data})")
        return VectorObject.from_dict(self.scene, result)

    def get_fill(self) -> GradientImageOrColor:
        return self.fill
    
    def set_fill(self, fill: GradientImageOrColor, recursive: bool) -> "VectorObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_fill(fill, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def get_fill_rule(self) -> str:
        return self.fill_rule
    
    def set_fill_rule(self, fill_rule: str, recursive: bool) -> "VectorObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_fill_rule(fill_rule, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def get_stroke(self) -> GradientImageOrColor:
        return self.stroke
    
    def set_stroke(self, stroke: GradientImageOrColor, recursive: bool) -> "VectorObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_stroke(stroke, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def get_stroke_width(self) -> float:
        return self.stroke_width
    
    def set_stroke_width(self, stroke_width: float, recursive: bool) -> "VectorObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_stroke_width(stroke_width, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=stroke_width,
            line_cap=self.line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def get_line_cap(self) -> str:
        return self.line_cap
    
    def set_line_cap(self, line_cap: str, recursive: bool) -> "VectorObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_line_cap(line_cap, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=line_cap,
            line_join=self.line_join,
            subobjects=subobjects,
            index=self.index,
        )
    
    def get_line_join(self) -> str:
        return self.line_join
    
    def set_line_join(self, line_join: str, recursive: bool) -> "VectorObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            subobjects = [subobject.set_line_join(line_join, recursive) for subobject in subobjects]
        return VectorObject(
            self.scene,
            points=self.points,
            fill=self.fill,
            fill_rule=self.fill_rule,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            line_cap=self.line_cap,
            line_join=line_join,
            subobjects=subobjects,
            index=self.index,
        )

    @staticmethod
    def from_dict(scene, data):
        return VectorObject(
            scene,
            points=[(point[0], point[1]) for point in data['points']],
            fill=GradientImageOrColor.from_dict(data['fill']) if data['fill'] else None,
            fill_rule=data['fillRule'],
            stroke=GradientImageOrColor.from_dict(data['stroke']) if data['stroke'] else None,
            stroke_width=data['strokeWidth'],
            line_cap=data['lineCap'],
            line_join=data['lineJoin'],
            subobjects=[VectorObject.from_dict(scene, subobject) for subobject in data['subobjects']],
            index=data['index'],
        )
    
    def to_dict(self):
        return {
            'type': 'vectorObject',
            'points': self.points,
            'fill': self.fill.to_dict(),
            'fillRule': self.fill_rule,
            'stroke': self.stroke.to_dict(),
            'strokeWidth': self.stroke_width,
            'lineCap': self.line_cap,
            'lineJoin': self.line_join,
            'subobjects': [subobject.to_dict() for subobject in self.subobjects],
            'index': self.index,
        }


class Color:
    def __init__(
        self,
        r: float,
        g: float,
        b: float,
        a: float = 1.0,
    ) -> None:
        self.r = r
        self.g = g
        self.b = b
        self.a = a
    
    @staticmethod
    def default() -> 'Color':
        return Color(0.0, 0.0, 0.0, 0.0)

    @staticmethod
    def from_dict(data) -> 'Color':
        return Color(data['r'], data['g'], data['b'], a=data['a'])

    def to_dict(self):
        return {
            'type': 'color',
            'r': self.r,
            'g': self.g,
            'b': self.b,
            'a': self.a,
        }


class GradientStop:
    def __init__(
        self,
        offset: float,
        color: Color,
    ) -> None:
        self.offset = offset
        self.color = color
    
    @staticmethod
    def from_dict(data) -> 'GradientStop':
        return GradientStop(data['offset'], Color.from_dict(data['color']))

    def to_dict(self):
        return {
            'type': 'gradientStop',
            'offset': self.offset,
            'color': self.color.to_dict(),
        }


class LinearGradient:
    def __init__(
        self,
        start: tuple[float, float],
        end: tuple[float, float],
        stops: list[GradientStop],
        alpha: float = 1.0,
    ) -> None:
        self.start = start
        self.end = end
        self.stops = stops
        self.alpha = alpha
    
    @staticmethod
    def from_dict(data) -> 'LinearGradient':
        return LinearGradient(
            (data['x1'], data['y1']),
            (data['x2'], data['y2']),
            [GradientStop.from_dict(stop) for stop in data['stops']],
            alpha=data['alpha'],
        )

    def to_dict(self):
        return {
            'type': 'linearGradient',
            'x1': self.start[0],
            'y1': self.start[1],
            'x2': self.end[0],
            'y2': self.end[1],
            'stops': [stop.to_dict() for stop in self.stops],
            'alpha': self.alpha,
        }


class RadialGradient:
    def __init__(
        self,
        c: tuple[float, float],
        r: float,
        f: tuple[float, float],
        stops: list[GradientStop],
        alpha: float = 1.0,
    ) -> None:
        self.c = c
        self.r = r
        self.f = f
        self.stops = stops
        self.alpha = alpha
    
    @staticmethod
    def from_dict(data) -> 'RadialGradient':
        return RadialGradient(
            (data['cx'], data['cy']),
            data['r'],
            (data['fx'], data['fy']),
            [GradientStop.from_dict(stop) for stop in data['stops']],
            alpha=data['alpha'],
        )

    def to_dict(self):
        return {
            'type': 'radialGradient',
            'cx': self.cx,
            'cy': self.cy,
            'r': self.r,
            'fx': self.fx,
            'fy': self.fy,
            'stops': [stop.to_dict() for stop in self.stops],
            'alpha': self.alpha,
        }


class Image:
    def __init__(
        self,
        image_base64: str,
        mime_type: str,
        top_left: tuple[float, float],
        bottom_right: tuple[float, float],
        alpha: float = 1.0,
    ) -> None:
        self.image_base64 = image_base64
        self.mime_type = mime_type
        self.top_left = top_left
        self.bottom_right = bottom_right
        self.alpha = alpha
    
    @staticmethod
    def from_dict(data) -> 'Image':
        return Image(
            data['imageBase64'],
            data['mimeType'],
            (data['left'], data['top']),
            (data['right'], data['bottom']),
            alpha=data['alpha'],
        )
    
    def to_dict(self):
        return {
            'type': 'image',
            'imageBase64': self.image_base64,
            'mimeType': self.mime_type,
            'left': self.top_left[0],
            'top': self.top_left[1],
            'right': self.bottom_right[0],
            'bottom': self.bottom_right[1],
            'alpha': self.alpha,
        }


class GradientImageOrColor:
    def __init__(
        self,
        gradient_image_or_color: LinearGradient | RadialGradient | Image | Color,
    ) -> None:
        self.gradient_image_or_color = gradient_image_or_color
    
    @staticmethod
    def default() -> 'GradientImageOrColor':
        return GradientImageOrColor(Color.default())
    
    @staticmethod
    def from_dict(data) -> 'GradientImageOrColor':
        if data['gradientImageOrColor']['type'] == 'linearGradient':
            return GradientImageOrColor(LinearGradient.from_dict(data['gradientImageOrColor']))
        elif data['gradientImageOrColor']['type'] == 'radialGradient':
            return GradientImageOrColor(RadialGradient.from_dict(data['gradientImageOrColor']))
        elif data['gradientImageOrColor']['type'] == 'image':
            return GradientImageOrColor(Image.from_dict(data['gradientImageOrColor']))
        elif data['gradientImageOrColor']['type'] == 'color':
            return GradientImageOrColor(Color.from_dict(data['gradientImageOrColor']))
        else:
            raise ValueError(f'Unknown type: {data["type"]}')
    
    def to_dict(self):
        return {
            'type': 'gradientImageOrColor',
            'gradientImageOrColor': self.gradient_image_or_color.to_dict(),
        }


from copy import deepcopy
import json
from typing import Callable

from .gradient_image_or_color import (
    GradientImageOrColor,
    Color,
    LinearGradient,
    RadialGradient,
    Image,
    GradientStop,
)
from .vector_object import VectorObject


class Camera:
    def __init__(
        self,
        scene,
        position: tuple[float, float, float],
        rotation: tuple[float, float, float],
        focal_distance: float,
        zoom: float,
    ) -> None:
        self.scene = scene
        self.position = position
        self.rotation = rotation
        self.focal_distance = focal_distance
        self.zoom = zoom
    
    async def project_points(self, points: list[tuple[float, float, float]]) -> list[tuple[float, float]]:
        data = json.dumps([[[point[0], point[1], point[2]] for point in points], self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'projectPoints', {data})", timeout=10)
        return [(point[0], point[1]) for point in result]
    
    @staticmethod
    def from_dict(data) -> 'Camera':
        return Camera(
            (data['position'][0], data['position'][1], data['position'][2]),
            (data['rotation'][0], data['rotation'][1], data['rotation'][2]),
            data['focalDistance'],
            data['zoom'],
        )
    
    def to_dict(self):
        return {
            'type': 'camera',
            'position': [self.position[0], self.position[1], self.position[2]],
            'rotation': [self.rotation[0], self.rotation[1], self.rotation[2]],
            'focalDistance': self.focal_distance,
            'zoom': self.zoom,
        }


class LightSource:
    def __init__(self, scene, position: tuple[float, float, float]) -> None:
        self.scene = scene
        self.position = position

    @staticmethod
    def from_dict(data) -> 'LightSource':
        return LightSource((data['position'][0], data['position'][1], data['position'][2]))
    
    def to_dict(self):
        return {
            'type': 'lightSource',
            'position': [self.position[0], self.position[1], self.position[2]],
        }


class ThreeDObject:
    def __init__(
        self,
        scene,
        points: list[tuple[float, float, float]] = None,
        subobjects: list['ThreeDObject'] = None,
        fill: GradientImageOrColor | None = None,
        stroke: GradientImageOrColor | None = None,
        stroke_width: float = 0.0,
        index: int = 0,
    ) -> None:
        points = points or []
        subobjects = subobjects or []
        fill = fill or GradientImageOrColor.default()
        stroke = stroke or GradientImageOrColor.default()
        self.scene = scene
        self.points = points
        self.subobjects = subobjects
        self.fill = fill
        self.stroke = stroke
        self.stroke_width = stroke_width
        self.index = index

    async def create_axes_3d(self, t: float, default_stroke_width: float = 0.5) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), t, default_stroke_width])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'createAxes3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def create(self, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'createThreeDObject', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def draw_stroke_then_fill(self, t: float, default_stroke_width: float = 0.5) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), t, default_stroke_width])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'drawStrokeThenFill3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def fade_in(self, scale_factor: float, shift: tuple[float, float, float], t: float) -> "ThreeDObject":
        shift = [shift[0], shift[1], shift[2]]
        data = json.dumps([self.to_dict(), scale_factor, shift, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'fadeIn3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def fade_out(self, scale_factor: float, shift: tuple[float, float, float], t: float) -> "ThreeDObject":
        shift = [shift[0], shift[1], shift[2]]
        data = json.dumps([self.to_dict(), scale_factor, shift, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'fadeOut3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def grow_from_center(self, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'growFromCenter3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def morph_shape(self, target: "ThreeDObject", t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), target.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'morphShape3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def rotate_x_animation(self, angle: float, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), angle, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotateXAnimation3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def rotate_y_animation(self, angle: float, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), angle, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotateYAnimation3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def rotate_z_animation(self, angle: float, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), angle, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotateZAnimation3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def scale_in_place(self, scale_factor: float, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), scale_factor, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'scaleInPlace3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def set_fill_animation(self, color: GradientImageOrColor, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), color.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'setFillAnimation3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def set_stroke_animation(self, color: GradientImageOrColor, t: float) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), color.to_dict(), t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'setStrokeAnimation3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def shift_animation(self, shift: tuple[float, float, float], t: float) -> "ThreeDObject":
        shift = [shift[0], shift[1], shift[2]]
        data = json.dumps([self.to_dict(), shift, t])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'shiftAnimation3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    def set_points(self, points: list[tuple[float, float, float]]) -> "ThreeDObject":
        return ThreeDObject(
            self.scene,
            points=points,
            subobjects=self.subobjects,
            fill=self.fill,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            index=self.index,
        )
    
    def set_subobjects(self, subobjects: list['ThreeDObject']) -> "ThreeDObject":
        return ThreeDObject(
            self.scene,
            points=self.points,
            subobjects=subobjects,
            fill=self.fill,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            index=self.index,
        )
    
    def get_points(self) -> list[tuple[float, float, float]]:
        return self.points
    
    def get_subobjects(self) -> list['ThreeDObject']:
        return self.subobjects
    
    def get_fill(self) -> GradientImageOrColor:
        return self.fill
    
    def get_stroke(self) -> GradientImageOrColor:
        return self.stroke
    
    def get_stroke_width(self) -> float:
        return self.stroke_width
    
    def get_index(self) -> int:
        return self.index
    
    def set_fill(self, fill: GradientImageOrColor, recursive: bool) -> "ThreeDObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            for i, subobject in enumerate(subobjects):
                subobjects[i] = subobject.set_fill(fill, recursive)
        return ThreeDObject(
            self.scene,
            points=self.points,
            subobjects=subobjects,
            fill=fill,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            index=self.index,
        )
    
    def set_stroke(self, stroke: GradientImageOrColor, recursive: bool) -> "ThreeDObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            for i, subobject in enumerate(subobjects):
                subobjects[i] = subobject.set_stroke(stroke, recursive)
        return ThreeDObject(
            self.scene,
            points=self.points,
            subobjects=subobjects,
            fill=self.fill,
            stroke=stroke,
            stroke_width=self.stroke_width,
            index=self.index,
        )
    
    def set_stroke_width(self, stroke_width: float, recursive: bool) -> "ThreeDObject":
        subobjects = deepcopy(self.subobjects)
        if recursive:
            for i, subobject in enumerate(subobjects):
                subobjects[i] = subobject.set_stroke_width(stroke_width, recursive)
        return ThreeDObject(
            self.scene,
            points=self.points,
            subobjects=self.subobjects,
            fill=self.fill,
            stroke=self.stroke,
            stroke_width=stroke_width,
            index=self.index,
        )
    
    def set_index(self, index: int) -> "ThreeDObject":
        return ThreeDObject(
            self.scene,
            points=self.points,
            subobjects=self.subobjects,
            fill=self.fill,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            index=index,
        )
    
    async def get_anchors_and_handles(self) -> tuple[list[tuple[float, float, float]], list[tuple[float, float, float]], list[tuple[float, float, float]], list[tuple[float, float, float]]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getAnchorsAndHandles3D', {data})")
        return (
            [(point[0], point[1], point[2]) for point in result[0]],
            [(point[0], point[1], point[2]) for point in result[1]],
            [(point[0], point[1], point[2]) for point in result[2]],
            [(point[0], point[1], point[2]) for point in result[3]],
        )
    
    async def set_anchors_and_handles(
        self,
        anchors_and_handles: tuple[list[tuple[float, float, float]], list[tuple[float, float, float]], list[tuple[float, float, float]], list[tuple[float, float, float]]],
    ) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), anchors_and_handles])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'setAnchorsAndHandles3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def scale_handle_to_anchor_distances(self, scale_factor: float, recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), scale_factor, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'scaleHandleToAnchorDistances3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def get_cubic_bezier_tuples(self) -> list[tuple[tuple[float, float, float], tuple[float, float, float], tuple[float, float, float], tuple[float, float, float]]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getCubicBezierTuples3D', {data})")
        return [
            (
                (point[0], point[1], point[2]),
                (point[3], point[4], point[5]),
                (point[6], point[7], point[8]),
                (point[9], point[10], point[11])
            )
            for point in result
        ]
    
    async def get_partial_copy(self, start: float, end: float, recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), start, end, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getPartialCopy3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    def set_fill_opacity(self, opacity: float, recursive: bool) -> "ThreeDObject":
        old_fill = self.fill.gradient_image_or_color
        if isinstance(old_fill, Color):
            new_fill = Color(old_fill.r, old_fill.g, old_fill.b, opacity)
        elif isinstance(old_fill, LinearGradient):
            stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in old_fill.stops
            ]
            new_fill = LinearGradient(old_fill.start, old_fill.end, stops, old_fill.alpha)
        elif isinstance(old_fill, RadialGradient):
            stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in old_fill.stops
            ]
            new_fill = RadialGradient(old_fill.center, old_fill.radius, stops, old_fill.alpha)
        elif isinstance(old_fill, Image):
            new_fill = Image(
                old_fill.image_base64,
                old_fill.mime_type,
                old_fill.top_left,
                old_fill.bottom_right,
                opacity,
            )
        subobjects = deepcopy(self.subobjects)
        if recursive:
            for i, subobject in enumerate(subobjects):
                subobjects[i] = subobject.set_fill_opacity(opacity, recursive)
        return ThreeDObject(
            self.scene,
            points=self.points,
            subobjects=subobjects,
            fill=GradientImageOrColor(new_fill),
            stroke=self.stroke,
            stroke_width=self.stroke_width,
            index=self.index,
        )
    
    def set_stroke_opacity(self, opacity: float, recursive: bool) -> "ThreeDObject":
        old_stroke = self.stroke.gradient_image_or_color
        if isinstance(old_stroke, Color):
            new_stroke = Color(old_stroke.r, old_stroke.g, old_stroke.b, opacity)
        elif isinstance(old_stroke, LinearGradient):
            stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in old_stroke.stops
            ]
            new_stroke = LinearGradient(old_stroke.start, old_stroke.end, stops, old_stroke.alpha)
        elif isinstance(old_stroke, RadialGradient):
            stops = [
                GradientStop(stop.offset, Color(stop.color.r, stop.color.g, stop.color.b, stop.color.a))
                for stop in old_stroke.stops
            ]
            new_stroke = RadialGradient(old_stroke.center, old_stroke.radius, stops, old_stroke.alpha)
        elif isinstance(old_stroke, Image):
            new_stroke = Image(
                old_stroke.image_base64,
                old_stroke.mime_type,
                old_stroke.top_left,
                old_stroke.bottom_right,
                opacity,
            )
        subobjects = deepcopy(self.subobjects)
        if recursive:
            for i, subobject in enumerate(subobjects):
                subobjects[i] = subobject.set_stroke_opacity(opacity, recursive)
        return ThreeDObject(
            self.scene,
            points=self.points,
            subobjects=subobjects,
            fill=self.fill,
            stroke=GradientImageOrColor(new_stroke),
            stroke_width=self.stroke_width,
            index=self.index,
        )
    
    def get_fill_opacity(self) -> float:
        return self.fill.gradient_image_or_color.a if isinstance(self.fill.gradient_image_or_color, Color) else self.fill.gradient_image_or_color.alpha
    
    def get_stroke_opacity(self) -> float:
        return self.stroke.gradient_image_or_color.a if isinstance(self.stroke.gradient_image_or_color, Color) else self.stroke.gradient_image_or_color.alpha
    
    async def get_critical_point(self, key: tuple[float, float, float]) -> tuple[float, float, float]:
        data = json.dumps([self.to_dict(), [key[0], key[1], key[2]]])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getCriticalPoint3D', {data})")
        return (result[0], result[1], result[2])
    
    async def next_to_other(
        self,
        other: "ThreeDObject",
        direction: tuple[float, float, float],
        buff: float,
        aligned_edge: tuple[float, float, float],
        recursive: bool,
    ) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), other.to_dict(), [direction[0], direction[1], direction[2]], buff, [aligned_edge[0], aligned_edge[1], aligned_edge[2]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'nextToOther3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def next_to_point(
        self,
        point: tuple[float, float, float],
        direction: tuple[float, float, float],
        buff: float,
        aligned_edge: tuple[float, float, float],
        recursive: bool,
    ) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), [point[0], point[1], point[2]], [direction[0], direction[1], direction[2]], buff, [aligned_edge[0], aligned_edge[1], aligned_edge[2]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'nextToPoint3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def arrange_subobjects(
        self,
        direction: tuple[float, float, float],
        buff: float,
        aligned_edge: tuple[float, float, float],
        recursive: bool,
    ) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), [direction[0], direction[1], direction[2]], buff, [aligned_edge[0], aligned_edge[1], aligned_edge[2]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'arrangeSubobjects3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def scale(self, scale_factor: float, recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), scale_factor, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'scale3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def strecth(self, scale_factor: tuple[float, float, float], recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), [scale_factor[0], scale_factor[1], scale_factor[2]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'stretch3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def shift(self, shift: tuple[float, float, float], recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), [shift[0], shift[1], shift[2]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'shift3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def rotate_x(self, angle: float, recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), angle, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotateX3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def rotate_y(self, angle: float, recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), angle, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotateY3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def rotate_z(self, angle: float, recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), angle, recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'rotateZ3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def project_and_shade(
        self,
        camera: Camera,
        light_source: LightSource,
    ) -> VectorObject:
        data = json.dumps([self.to_dict(), camera.to_dict(), light_source.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'projectAndShade3D', {data})", timeout=10)
        return VectorObject.from_dict(self.scene, result)
    
    async def apply_function(
        self,
        func: Callable[[tuple[float, float, float]], tuple[float, float, float]],
        recursive: bool,
    ) -> "ThreeDObject":
        id_ = self.scene.register_callback(func)
        data = json.dumps([self.to_dict(), id_, recursive])
        result = await self.scene.client.run_javascript(f"return await runMethod({self.scene.id}, 'applyFunction3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def get_bounding_box(self) -> tuple[tuple[float, float, float], tuple[float, float, float]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getBoundingBox3D', {data})")
        return (
            (result[0][0], result[0][1], result[0][2]),
            (result[1][0], result[1][1], result[1][2]),
        )
    
    async def get_center(self) -> tuple[float, float, float]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getCenter3D', {data})")
        return (result[0], result[1], result[2])
    
    async def merged_points(self) -> list[tuple[float, float, float]]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'mergedPoints3D', {data})")
        return [(point[0], point[1], point[2]) for point in result]
    
    async def move_to(self, point: tuple[float, float, float], recursive: bool) -> "ThreeDObject":
        data = json.dumps([self.to_dict(), [point[0], point[1], point[2]], recursive])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'moveTo3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def get_subobjects_recursively(self) -> list["ThreeDObject"]:
        data = json.dumps([self.to_dict()])
        result = await self.scene.client.run_javascript(f"return runMethod({self.scene.id}, 'getSubobjectsRecursively3D', {data})")
        return [ThreeDObject.from_dict(self.scene, subobject) for subobject in result]

    @staticmethod
    def from_dict(scene, data):
        return ThreeDObject(
            scene,
            points=[(point[0], point[1], point[2]) for point in data['points']],
            subobjects=[ThreeDObject.from_dict(scene, subobject) for subobject in data['subobjects']],
            fill=GradientImageOrColor.from_dict(data['fill']) if data['fill'] else None,
            stroke=GradientImageOrColor.from_dict(data['stroke']) if data['stroke'] else None,
            stroke_width=data['strokeWidth'],
            index=data['index'],
        )
    
    def to_dict(self):
        return {
            'type': 'threeDObject',
            'points': self.points,
            'subobjects': [subobject.to_dict() for subobject in self.subobjects],
            'fill': self.fill.to_dict(),
            'stroke': self.stroke.to_dict(),
            'strokeWidth': self.stroke_width,
            'index': self.index,
        }


from asyncio import Event, iscoroutine
import json
from pathlib import Path
from typing import Callable, Coroutine, Any

from nicegui.element import Element
from nicegui.dependencies import libraries, Library
from nicegui.events import GenericEventArguments

import mathlikeanim_rs
from .gradient_image_or_color import GradientImageOrColor, Color
from .three_d_object import ThreeDObject, Camera, LightSource
from .vector_object import VectorObject


libraries['mathlikeanim-rs'] = Library(
    key='mathlikeanim-rs',
    name='mathlikeanim-rs',
    path=Path(mathlikeanim_rs.__file__).parent / 'node_modules' / 'mathlikeanim-rs' / 'index.js',
    expose=True,
)
libraries['mathlikeanim-rs/index_bg.wasm'] = Library(
    key='mathlikeanim-rs/index_bg.wasm',
    name='mathlikeanim-rs/index_bg.wasm',
    path=Path(mathlikeanim_rs.__file__).parent / 'node_modules' / 'mathlikeanim-rs' / 'index_bg.wasm',
    expose=True,
)


class Scene(
    Element,
    component='scene.js',
):
    def __init__(
        self,
        width: int = 1920,
        height: int = 1080,
        css_width: str = '50vw',
        css_height: str = 'auto',
        fps: int = 60,
        svg: bool = False,
    ) -> None:
        super().__init__()
        self._props['width'] = width
        self._props['height'] = height
        self._props['cssWidth'] = css_width
        self._props['cssHeight'] = css_height
        self._props['fps'] = fps
        self._props['svg'] = svg
        self._ready = Event()
        self._funcs = {}
        self.on('ready', self._on_ready)
        self.on('python-request', self._on_python_request, ['pythonFuncId', 'args'])
    
    async def _on_python_request(self, data: GenericEventArguments):
        data = data.args
        id_ = data['pythonFuncId']
        func = self._funcs[id_]
        result = func(*data['args'])
        if iscoroutine(result):
            result = await result
        data = json.dumps([result])
        self.client.run_javascript(f'return runMethod({self.id}, "emitPythonResponse", {data})')

    async def wait_until_ready(self):
        await self._ready.wait()
    
    def _on_ready(self):
        self._ready.set()

    def new_empty_object(self) -> VectorObject:
        return VectorObject(self)
    
    def new_empty_three_d_object(self) -> ThreeDObject:
        return ThreeDObject(self)
    
    async def add_final_tip_to_object(
        self,
        shape: VectorObject,
        tip_side_length: float = 20,
        tip_color: Color | None = None,
    ) -> VectorObject:
        tip_color = tip_color or Color(1, 1, 1, a=1)
        data = json.dumps([shape.to_dict(), tip_side_length, tip_color.to_dict()])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "addFinalTipToObject", {data})')
        return VectorObject.from_dict(self, result)

    async def add_initial_tip_to_object(
        self,
        shape: VectorObject,
        tip_side_length: float = 20,
        tip_color: Color | None = None,
    ) -> VectorObject:
        tip_color = tip_color or Color(1, 1, 1, a=1)
        data = json.dumps([shape.to_dict(), tip_side_length, tip_color.to_dict()])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "addInitialTipToObject", {data})')
        return VectorObject.from_dict(self, result)

    async def add_both_sides_tips_to_object(
        self,
        shape: VectorObject,
        tip_side_length: float = 20,
        tip_color: Color | None = None,
    ) -> VectorObject:
        tip_color = tip_color or Color(1, 1, 1, a=1)
        data = json.dumps([shape.to_dict(), tip_side_length, tip_color.to_dict()])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "addBothSidesTipsToObject", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_arc(
        self,
        center: tuple[float, float],
        radius: float,
        start_angle: float,
        end_angle: float,
        n_samples: int = 10,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], radius, start_angle, end_angle, n_samples])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newArc", {data})')
        return VectorObject.from_dict(self, result)

    async def new_circle(
        self,
        center: tuple[float, float],
        radius: float,
        num_points: int = 10,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], radius, num_points])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newCircle", {data})')
        return VectorObject.from_dict(self, result)

    async def new_elliptical_arc(
        self,
        center: tuple[float, float],
        radius_x: float,
        radius_y: float,
        start_angle: float,
        end_angle: float,
        n_samples: int = 10,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], radius_x, radius_y, start_angle, end_angle, n_samples])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newEllipticalArc", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_ellipse(
        self,
        center: tuple[float, float],
        radius_x: float,
        radius_y: float,
        num_points: int = 10,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], radius_x, radius_y, num_points])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newEllipse", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_annular_sector(
        self,
        center: tuple[float, float],
        inner_radius: float,
        outer_radius: float,
        start_angle: float,
        end_angle: float,
        n_samples: int = 10,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], inner_radius, outer_radius, start_angle, end_angle, n_samples])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newAnnularSector", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_dashed_object(
        self,
        vector_object: VectorObject,
        num_dashes: int = 15,
        dashed_ratio: float = 0.5,
        dash_offset: float = 0.0,
        equal_length: bool = True,
    ) -> VectorObject:
        data = json.dumps([vector_object.to_dict(), num_dashes, dashed_ratio, dash_offset, equal_length])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newDashedObject", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_line(
        self,
        start: tuple[float, float],
        end: tuple[float, float],
        color: Color | None = None,
        stroke_width: float = 4.0,
    ) -> VectorObject:
        color = color or Color(1, 1, 1, a=1)
        data = json.dumps([[start[0], start[1]], [end[0], end[1]], color.to_dict(), stroke_width])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newLine", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_polygon(
        self,
        points: list[tuple[float, float]],
    ) -> VectorObject:
        data = json.dumps([[[point[0], point[1]] for point in points]])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newPolygon", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_regular_polygon(
        self,
        center: tuple[float, float],
        side_length: float,
        num_sides: int,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], side_length, num_sides])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newRegularPolygon", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_square(
        self,
        center: tuple[float, float],
        side_length: float,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], side_length])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newSquare", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_rectangle(
        self,
        center: tuple[float, float],
        width: float,
        height: float,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], width, height])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newRectangle", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_equilateral_triangle(
        self,
        center: tuple[float, float],
        side_length: float,
    ) -> VectorObject:
        data = json.dumps([[center[0], center[1]], side_length])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newEquilateralTriangle", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_triangle(
        self,
        point1: tuple[float, float],
        point2: tuple[float, float],
        point3: tuple[float, float],
    ) -> VectorObject:
        data = json.dumps([[point1[0], point1[1]], [point2[0], point2[1]], [point3[0], point3[1]]])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newTriangle", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_right_triangle(
        self,
        point1: tuple[float, float],
        point2: tuple[float, float],
    ) -> VectorObject:
        data = json.dumps([[point1[0], point1[1]], [point2[0], point2[1]]])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newRightTriangle", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_axes(
        self,
        x_range: tuple[float, float, float],
        y_range: tuple[float, float, float],
        center: tuple[float, float],
        x_length: float = 1000.0,
        y_length: float = 1000.0,
        color: Color | None = None,
        stroke_width: float = 4.0,
        line_cap: str = 'butt',
        line_join: str = 'miter',
        index: int = 0,
        add_x_ticks: bool = True,
        add_y_ticks: bool = True,
        x_tick_size: float = 20.0,
        y_tick_size: float = 20.0,
        add_x_tip: bool = True,
        add_y_tip: bool = True,
    ) -> VectorObject:
        color = color or Color(1, 1, 1, a=1)
        x_min, x_max, x_step = x_range
        y_min, y_max, y_step = y_range
        data = json.dumps([
            x_min, x_max, x_step,
            y_min, y_max, y_step,
            [center[0], center[1]],
            x_length, y_length,
            color.to_dict(),
            stroke_width,
            line_cap, line_join,
            index,
            add_x_ticks, add_y_ticks,
            x_tick_size, y_tick_size,
            add_x_tip, add_y_tip,
        ])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newAxes", {data})')
        return VectorObject.from_dict(self, result)

    async def axes_coords_to_point(
        self,
        axes: VectorObject,
        x: float,
        y: float,
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
    ) -> tuple[float, float]:
        data = json.dumps([axes.to_dict(), x, y, x_min, x_max, y_min, y_max])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "axesCoordsToPoint", {data})')
        return tuple(result)
    
    async def point_to_axes_coords(
        self,
        axes: VectorObject,
        point: tuple[float, float],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
    ) -> tuple[float, float]:
        data = json.dumps([axes.to_dict(), [point[0], point[1]], x_min, x_max, y_min, y_max])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "pointToAxesCoords", {data})')
        return tuple(result)
    
    async def new_parametric_plot_in_axes(
        self,
        f: Callable[[float], tuple[float, float]],
        t_range: tuple[float, float, float],
        axes: VectorObject,
        xMin: float,
        xMax: float,
        yMin: float,
        yMax: float,
    ) -> VectorObject:
        t_min, t_max, t_step = t_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, t_min, t_max, t_step, axes.to_dict(), xMin, xMax, yMin, yMax])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newParametricPlotInAxes", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_plot_in_axes(
        self,
        f: Callable[[float], float],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        x_range: tuple[float, float, float],
        axes: VectorObject,
    ) -> VectorObject:
        x_1, x_2, x_step = x_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, x_min, x_max, y_min, y_max, x_1, x_2, x_step, axes.to_dict()])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newPlotInAxes", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_contour_plot_in_axes(
        self,
        f: Callable[[float, float], float],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        x_range: tuple[float, float, float],
        y_range: tuple[float, float, float],
        axes: VectorObject,
        intervals: list[float],
    ) -> VectorObject:
        x_1, x_2, x_step = x_range
        y_1, y_2, y_step = y_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, x_min, x_max, y_min, y_max, x_1, x_2, x_step, y_1, y_2, y_step, axes.to_dict(), intervals])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newContourPlotInAxes", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_area_under_curve(
        self,
        axes: VectorObject,
        plot: VectorObject,
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        x_range: tuple[float, float],
    ) -> VectorObject: 
        x_1, x_2 = x_range
        data = json.dumps([axes.to_dict(), plot.to_dict(), x_min, x_max, y_min, y_max, x_1, x_2])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newAreaUnderCurve", {data})')
        return VectorObject.from_dict(self, result)

    async def riemann_rectangles_for_plot(
        self,
        f: Callable[[float], float],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        direction: float,
        x_range: tuple[float, float],
        n_rects: int,
        axes: VectorObject,
    ) -> VectorObject:
        x_1, x_2 = x_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, x_min, x_max, y_min, y_max, direction, x_1, x_2, n_rects, axes.to_dict()])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "riemannRectanglesForPlot", {data})')
        return VectorObject.from_dict(self, result)

    async def secant_line_for_plot(
        self,
        f: Callable[[float], float],
        x_range: tuple[float, float],
        length: float,
        axes: VectorObject,
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
    ) -> VectorObject:
        x_1, x_2 = x_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, x_1, x_2, length, axes.to_dict(), x_min, x_max, y_min, y_max])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "secantLineForPlot", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_parametric_plot(
        self,
        f: Callable[[float], tuple[float, float]],
        t_range: tuple[float, float, float],
    ) -> VectorObject:
        t_min, t_max, t_step = t_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, t_min, t_max, t_step])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newParametricPlot", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_real_function(
        self,
        f: Callable[[float], float],
        x_range: tuple[float, float, float],
    ) -> VectorObject:
        x_min, x_max, x_step = x_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, x_min, x_max, x_step])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newRealFunction", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_contour_plot(
        self,
        f: Callable[[float, float], float],
        x_range: tuple[float, float, float],
        y_range: tuple[float, float, float],
        intervals: list[float],
    ) -> VectorObject:
        x_min, x_max, x_step = x_range
        y_min, y_max, y_step = y_range
        id_ = self.register_callback(f)
        data = json.dumps([id_, x_min, x_max, x_step, y_min, y_max, y_step, intervals])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newContourPlot", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_number_line(
        self,
        x_range: tuple[float, float, float],
        center: tuple[float, float],
        color: Color | None = None,
        stroke_width: float = 4.0,
        line_cap: str = 'butt',
        line_join: str = 'miter',
        index: int = 0,
        length: float = 1000.0,
        add_tip: bool = True,
        add_ticks: bool = True,
        tick_size: float = 20.0,
        angle: float = 0.0,
    ) -> VectorObject:
        color = color or Color(1, 1, 1, a=1)
        x_min, x_max, x_step = x_range
        data = json.dumps([
            x_min, x_max, x_step,
            [center[0], center[1]],
            color.to_dict(),
            stroke_width,
            line_cap, line_join,
            index,
            length,
            add_tip,
            add_ticks,
            tick_size,
            angle,
        ])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newNumberLine", {data})')
        return VectorObject.from_dict(self, result)
    
    async def number_to_point_on_number_line(
        self,
        number_line: VectorObject,
        number: float,
        x_min: float,
        x_max: float,
    ) -> tuple[float, float]:
        data = json.dumps([number_line.to_dict(), number, x_min, x_max])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "numberToPointOnNumberLine", {data})')
        return tuple(result)
    
    async def point_on_number_line_to_number(
        self,
        number_line: VectorObject,
        point: tuple[float, float],
        x_min: float,
        x_max: float,
    ) -> float:
        data = json.dumps([number_line.to_dict(), [point[0], point[1]], x_min, x_max])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "pointOnNumberLineToNumber", {data})')
        return result
    
    async def get_numbers_tex(
        self,
        number_line: VectorObject,
        numbers: list[float],
        number_to_vector_object: Callable[[float], Coroutine[None, None, VectorObject]],
        x_min: float,
        x_max: float,
        height: float,
        direction: tuple[float, float] | None = None,
        buff: float = 20.0,
        index: int = 0,
    ) -> VectorObject:
        direction = direction or (0, 1)
        direction = [direction[0], direction[1]]
        async def number_to_vector_object_async(number):
            result = await number_to_vector_object(number)
            return result.to_dict()
        id_ = self.register_callback(number_to_vector_object_async)
        data = json.dumps([number_line.to_dict(), numbers, id_, x_min, x_max, height, direction, buff, index])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "getNumbersTex", {data})')
        return VectorObject.from_dict(self, result)

    async def new_mathjax(
        self,
        expression: str,
        font_base64: dict[str, str] | None = None,
    ) -> VectorObject:
        font_base64 = font_base64 or {}
        data = json.dumps([expression, font_base64])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newMathjax", {data})')
        return VectorObject.from_dict(self, result)
    
    async def svg_to_vector_object(
        self,
        svg: str,
        font_base64: dict[str, str] | None = None,
    ) -> VectorObject:
        font_base64 = font_base64 or {}
        data = json.dumps([svg, font_base64])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "svgToVectorObject", {data})')
        return VectorObject.from_dict(self, result)
    
    async def text_to_vector_object(
        self,
        text: str,
        font_base64: dict[str, str] | None = None,
        font_weight: str = 'normal',
        font_style: str = 'normal',
        x: float = 0.0,
        y: float = 0.0,
        font_size: float = 20.0,
    ) -> VectorObject:
        font_base64 = font_base64 or {}
        data = json.dumps([text, font_base64, font_weight, font_style, x, y, font_size])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "textToVectorObject", {data})')
        return VectorObject.from_dict(self, result)
    
    async def new_sphere(
        self,
        center: tuple[float, float, float],
        radius: float,
        u_segments: int,
        v_segments: int,
        fill_colors: list[Color],
        stroke_colors: list[Color],
        stroke_width: float,
        index: int,
    ) -> ThreeDObject:
        data = json.dumps([
            [center[0], center[1], center[2]],
            radius,
            u_segments,
            v_segments,
            [color.to_dict() for color in fill_colors],
            [color.to_dict() for color in stroke_colors],
            stroke_width,
            index,
        ])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newSphere", {data})', timeout=60)
        return ThreeDObject.from_dict(self, result)
    
    async def new_three_d_axes(
        self,
        x_range: tuple[float, float, float],
        y_range: tuple[float, float, float],
        z_range: tuple[float, float, float],
        center: tuple[float, float, float],
        x_length: float = 1000.0,
        y_length: float = 1000.0,
        z_length: float = 1000.0,
        color: Color | None = None,
        stroke_width: float = 4.0,
        add_x_ticks: bool = True,
        add_y_ticks: bool = True,
        add_z_ticks: bool = True,
        x_tick_size: float = 20.0,
        y_tick_size: float = 20.0,
        z_tick_size: float = 20.0,
        add_x_tip: bool = True,
        add_y_tip: bool = True,
        add_z_tip: bool = True,
        n_pieces: int = 20,
        index: int = 0,
    ) -> ThreeDObject:
        color = color or Color(1, 1, 1, a=1)
        x_min, x_max, x_step = x_range
        y_min, y_max, y_step = y_range
        z_min, z_max, z_step = z_range
        data = json.dumps([
            x_min, x_max, x_step,
            y_min, y_max, y_step,
            z_min, z_max, z_step,
            [center[0], center[1], center[2]],
            x_length, y_length, z_length,
            color.to_dict(),
            stroke_width,
            add_x_ticks, add_y_ticks, add_z_ticks,
            x_tick_size, y_tick_size, z_tick_size,
            add_x_tip, add_y_tip, add_z_tip,
            n_pieces,
            index,
        ])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "newThreeDAxes", {data})')
        return ThreeDObject.from_dict(self, result)
    
    async def coords_to_point_on_three_d_axes(
        self,
        axes: ThreeDObject,
        coords: tuple[float, float, float],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        z_min: float,
        z_max: float,
    ) -> tuple[float, float, float]:
        coords = [coords[0], coords[1], coords[2]]
        data = json.dumps([axes.to_dict(), coords, x_min, x_max, y_min, y_max, z_min, z_max])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "coordsToPointOnThreeDAxes", {data})')
        return tuple(result)
    
    async def point_on_three_d_axes_to_coords(
        self,
        axes: ThreeDObject,
        point: tuple[float, float, float],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        z_min: float,
        z_max: float,
    ) -> tuple[float, float, float]:
        point = [point[0], point[1], point[2]]
        data = json.dumps([axes.to_dict(), point, x_min, x_max, y_min, y_max, z_min, z_max])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "pointOnThreeDAxesToCoords", {data})')
        return tuple(result)
    
    async def new_parametric_plot_in_three_d_axes(
        self,
        axes: ThreeDObject,
        f: Callable[[float, float], tuple[float, float, float]],
        u_range: tuple[float, float, int],
        v_range: tuple[float, float, int],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        z_min: float,
        z_max: float,
        fills: list[Color],
        strokes: list[Color],
        stroke_width: float,
        index: int = 0,
    ) -> ThreeDObject:
        u_min, u_max, u_steps = u_range
        v_min, v_max, v_steps = v_range
        id_ = self.register_callback(f)
        data = json.dumps([
            axes.to_dict(),
            id_,
            u_min, u_max,
            v_min, v_max,
            u_steps, v_steps,
            x_min, x_max,
            y_min, y_max,
            z_min, z_max,
            [color.to_dict() for color in fills],
            [color.to_dict() for color in strokes],
            stroke_width,
            index])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newParametricPlotInThreeDAxes", {data})')
        return ThreeDObject.from_dict(self, result)
    
    async def new_plot_in_three_d_axes(
        self,
        axes: ThreeDObject,
        f: Callable[[float, float], float],
        u_range: tuple[float, float, int],
        v_range: tuple[float, float, int],
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        z_min: float,
        z_max: float,
        fills: list[Color],
        strokes: list[Color],
        stroke_width: float,
        index: int = 0,
    ) -> ThreeDObject:
        u_min, u_max, u_steps = u_range
        v_min, v_max, v_steps = v_range
        id_ = self.register_callback(f)
        data = json.dumps([
            axes.to_dict(),
            id_,
            u_min, u_max,
            v_min, v_max,
            u_steps, v_steps,
            x_min, x_max,
            y_min, y_max,
            z_min, z_max,
            [color.to_dict() for color in fills],
            [color.to_dict() for color in strokes],
            stroke_width,
            index,
        ])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newPlotInThreeDAxes", {data})')
        return ThreeDObject.from_dict(self, result)
    
    async def new_parametric_line_plot_in_three_d_axes(
        self,
        axes: ThreeDObject,
        f: Callable[[float], tuple[float, float, float]],
        t_range: tuple[float, float, int],     
        x_min: float,
        x_max: float,
        y_min: float,
        y_max: float,
        z_min: float,
        z_max: float,   
        color: Color,
        stroke_width: float,
        index: int = 0,
    ) -> ThreeDObject:
        t_min, t_max, t_steps = t_range
        id_ = self.register_callback(f)
        data = json.dumps([
            axes.to_dict(),
            id_,
            t_min, t_max, t_steps,
            x_min, x_max,
            y_min, y_max,
            z_min, z_max,
            color.to_dict(),
            stroke_width,
            index,
        ])
        result = await self.client.run_javascript(f'return await runMethod({self.id}, "newParametricLinePlotInThreeDAxes", {data})')
        return ThreeDObject.from_dict(self, result)

    async def from_uv_function(
        self,
        func: Callable[[tuple[float, float]], tuple[float, float, float]],
        u_range: tuple[float, float],
        v_range: tuple[float, float],
        u_num_steps: int,
        v_num_steps: int,
        fills: list[Color],
        strokes: list[Color],
        stroke_width: float,
        index: int = 0,
    ) -> "ThreeDObject":
        id_ = self.register_callback(func)
        data = json.dumps([id_, u_range, v_range, u_num_steps, v_num_steps, [fill.to_dict() for fill in fills], [stroke.to_dict() for stroke in strokes], stroke_width, index])
        result = await self.client.run_javascript(f"return await runMethod({self.scene.id}, 'fromUVFunction3D', {data})")
        return ThreeDObject.from_dict(self.scene, result)
    
    async def three_d_object_from_vector_object(
        self,
        vector_object: VectorObject,
    ) -> ThreeDObject:
        data = json.dumps([vector_object.to_dict()])
        result = await self.client.run_javascript(f"return await runMethod({self.id}, 'threeDObjectFromVectorObject', {data})")
        return ThreeDObject.from_dict(self, result)
    
    async def add(self, vector_object: VectorObject) -> None:
        data = json.dumps([vector_object.to_dict()])
        await self.client.run_javascript(f'return runMethod({self.id}, "addToScene", {data})')

    async def remove(self, index: int) -> None:
        await self.client.run_javascript(f'return runMethod({self.id}, "removeFromScene", {index})')

    async def clear(self) -> None:
        data = json.dumps([])
        await self.client.run_javascript(f'return runMethod({self.id}, "clearScene", {data})')
    
    async def restore(self, index: int) -> None:
        data = json.dumps([index])
        await self.client.run_javascript(f'return runMethod({self.id}, "restoreState", {data})')
    
    async def save_state(self, index: int) -> None:
        data = json.dumps([index])
        await self.client.run_javascript(f'return runMethod({self.id}, "saveState", {data})')

    async def set_top_left_corner(self, x: float, y: float) -> None:
        data = json.dumps([x, y])
        await self.client.run_javascript(f'return runMethod({self.id}, "setTopLeftCorner", {data})')

    async def set_bottom_right_corner(self, x: float, y: float) -> None:
        data = json.dumps([x, y])
        await self.client.run_javascript(f'return runMethod({self.id}, "setBottomRightCorner", {data})')
    
    async def set_background(self, background: GradientImageOrColor) -> None:
        data = json.dumps([background.to_dict()])
        await self.client.run_javascript(f'return runMethod({self.id}, "setBackground", {data})')

    async def play(
        self,
        animation_func: Callable[[dict[int, VectorObject], float], Coroutine[None, None, dict[int, VectorObject]]],
        object_indices: list[int],
        duration_in_frames: int,
    ) -> None:
        async def animation_func_async(objects, time):
            result = await animation_func({int(index): VectorObject.from_dict(self, obj) for index, obj in objects.items()}, time)
            return {str(index): obj.to_dict() for index, obj in result.items()}
        id_ = self.register_callback(animation_func_async)
        data = json.dumps([id_, object_indices, duration_in_frames])
        await self.client.run_javascript(f'return await runMethod({self.id}, "play", {data})', timeout=2.5*duration_in_frames / self._props['fps'])
    
    async def set_updater(self, index: int, updater: Callable[[VectorObject], VectorObject]) -> None:
        id_ = self.register_callback(lambda obj: updater(VectorObject.from_dict(self, obj)).to_dict())
        data = json.dumps([index, id_])
        await self.client.run_javascript(f'return runMethod({self.id}, "setUpdater", {data})')
    
    async def remove_updater(self, index: int) -> None:
        data = json.dumps([index])
        await self.client.run_javascript(f'return runMethod({self.id}, "removeUpdater", {data})')
    
    async def wait(self, duration_in_frames: int, object_indices: list[int]) -> None:
        data = json.dumps([duration_in_frames, object_indices])
        await self.client.run_javascript(f'return await runMethod({self.id}, "wait", {data})', timeout=2.5*duration_in_frames / self._props['fps'])
    
    async def wait_until(
        self,
        condition: Callable[[], Coroutine[None, None, bool]],
        object_indices: list[int],
    ) -> None:
        id_ = self.register_callback(condition)
        data = json.dumps([id_, object_indices])
        await self.client.run_javascript(f'return await runMethod({self.id}, "waitUntil", {data})', timeout=10e9)
    
    async def set_on_rendered(self, func: Callable[[], Coroutine[None, None, None]]) -> None:
        id_ = self.register_callback(func)
        data = json.dumps([id_])
        await self.client.run_javascript(f'return runMethod({self.id}, "setOnRendered", {data})')
    
    async def render_frame(self) -> None:
        data = json.dumps([])
        await self.client.run_javascript(f'return await runMethod({self.id}, "renderFrame", {data})')
    
    async def get_objects(self) -> list[VectorObject]:
        data = json.dumps([])
        result = await self.client.run_javascript(f'return runMethod({self.id}, "getObjects", {data})')
        return [VectorObject.from_dict(self, obj) for obj in result]
    
    def new_camera(
        self,
        position: tuple[float, float, float],
        rotation: tuple[float, float, float],
        focal_distance: float,
        zoom: float,
    ) -> Camera:
        return Camera(self, position, rotation, focal_distance, zoom)
    
    def new_light_source(
        self,
        position: tuple[float, float, float],
    ) -> LightSource:
        return LightSource(self, position)

    async def exec_js(self, js: str) -> Any:
        return await self.client.run_javascript(js, timeout=10e9)
    
    async def log(self, message: str) -> None:
        data = json.dumps([message])
        await self.client.run_javascript(f"return runMethod({self.id}, 'log', {data})")

    def register_callback(self, func: Callable):
        id_ = id(func)
        self._funcs[id_] = func
        return id_